//
// Created by maxine on 7/1/24.
//

#include <mcolor.h>
#include <format>
#include <iostream>

#ifdef WIN32
#include <windows.h>
#endif

namespace mcolor
{
    std::string toEscapeCode(Color4 c, bool bg)
    {
        if (bg)
            return std::format("\033[48;2;{};{};{}m", c.r, c.g, c.b);

        return std::format("\033[38;2;{};{};{}m", c.r, c.g, c.b);
    }

    std::string toEscapeCode(AnsiColor c)
    {
        // Type casting is annoying just saying, but I want to be special about this for some reason
        return std::format("\033[{}m", static_cast<typename std::underlying_type<AnsiColor>::type>(c));
    }

#ifdef WIN32
    /*
     * Beat Windows into submission and make it interpret ansi codes.
     * Fuck you Microsoft we're doing this the right way.
     * This only works on Windows 10 version 1511 and newer.
     *
     * This function is meant to be a helper that's used in ifdef macros to allow
     * us to print ansi color codes to the terminal. It's always gonna have to tag
     * along with any code that is printing the escape codes to a terminal.
     */
    void windowsSaneify() {
            HANDLE handleOut = GetStdHandle(STD_OUTPUT_HANDLE);
            DWORD consoleMode;
            GetConsoleMode( handleOut, &consoleMode);
            consoleMode |= ENABLE_VIRTUAL_TERMINAL_PROCESSING;
            consoleMode |= DISABLE_NEWLINE_AUTO_RETURN;
            SetConsoleMode( handleOut , consoleMode );
            SetConsoleOutputCP(CP_UTF8);
    }
#endif

    void printAnsiColorTable() {
        std::vector<AnsiColor> ansifg = {
                AnsiColor::FG_BLACK,
                AnsiColor::FG_RED,
                AnsiColor::FG_GREEN,
                AnsiColor::FG_YELLOW,
                AnsiColor::FG_BLUE,
                AnsiColor::FG_MAGENTA,
                AnsiColor::FG_CYAN,
                AnsiColor::FG_WHITE,
        };

        std::vector<AnsiColor> ansibg = {
                AnsiColor::BG_BLACK,
                AnsiColor::BG_RED,
                AnsiColor::BG_GREEN,
                AnsiColor::BG_YELLOW,
                AnsiColor::BG_BLUE,
                AnsiColor::BG_MAGENTA,
                AnsiColor::BG_CYAN,
                AnsiColor::BG_WHITE,
        };

        std::vector<AnsiColor> ansifg_bright = {
                AnsiColor::FG_BRIGHT_BLACK,
                AnsiColor::FG_BRIGHT_RED,
                AnsiColor::FG_BRIGHT_GREEN,
                AnsiColor::FG_BRIGHT_YELLOW,
                AnsiColor::FG_BRIGHT_BLUE,
                AnsiColor::FG_BRIGHT_MAGENTA,
                AnsiColor::FG_BRIGHT_CYAN,
                AnsiColor::FG_BRIGHT_WHITE,
        };

        std::vector<AnsiColor> ansibg_bright = {
                AnsiColor::BG_BRIGHT_BLACK,
                AnsiColor::BG_BRIGHT_RED,
                AnsiColor::BG_BRIGHT_GREEN,
                AnsiColor::BG_BRIGHT_YELLOW,
                AnsiColor::BG_BRIGHT_BLUE,
                AnsiColor::BG_BRIGHT_MAGENTA,
                AnsiColor::BG_BRIGHT_CYAN,
                AnsiColor::BG_BRIGHT_WHITE,
        };

        std::vector<AnsiColor> all;
        all.insert(all.end(), ansifg.begin(), ansifg.end());
        all.insert(all.end(), ansibg.begin(), ansibg.end());
        all.insert(all.end(), ansifg_bright.begin(), ansifg_bright.end());
        all.insert(all.end(), ansibg_bright.begin(), ansibg_bright.end());

        for (const auto i : all)
        {
            auto n = static_cast<typename std::underlying_type<AnsiColor>::type>(i);
            std::cout << mcolor::toEscapeCode(i) << unsigned(n) << mcolor::toEscapeCode(AnsiColor::RESET);

            if (n == 37 || n == 47 || n == 97 || n == 107)
                std::cout << std::endl;
        }
    }

    std::vector<uint8_t> createRGBScale(uint8_t startValue, uint8_t increment)
    {
        std::vector<uint8_t> scale;
        for (int i = startValue; i < 256; i += increment)
        {
            scale.push_back(i);
        }
        return scale;
    }

    std::vector<Color4> redScaler(Color4 c)
    {
        std::vector<Color4> scale;
        auto nscale = createRGBScale(c.r);
        for (auto const i : nscale)
        {
            scale.push_back(Color4(i, c.g, c.b));
        }
        return scale;
    }

    std::vector<Color4> greenScaler(Color4 c)
    {
        std::vector<Color4> scale;
        auto nscale = createRGBScale(c.g);
        for (auto const i : nscale)
        {
            scale.push_back(Color4(c.r, i, c.b));
        }
        return scale;
    }

    std::vector<Color4> blueScaler(Color4 c)
    {
        std::vector<Color4> scale;
        auto nscale = createRGBScale(c.b);
        for (auto const i : nscale)
        {
            scale.push_back(Color4(c.r, c.g, i));
        }
        return scale;
    }

    std::vector<Color4> RGBColorScale(Color4 startingColor, std::vector<Color4> (*scaler)(Color4))
    {
        return scaler(startingColor);
    }

    void printRGBScale(std::vector<Color4> cs)
    {
        for (auto i : cs) {
            std::cout << mcolor::toEscapeCode(i) << mcolor::toEscapeCode(i, true) << ":3" << mcolor::toEscapeCode(AnsiColor::RESET);
        }
        std::cout << std::endl;
    }

    /*
     * Some terminals/consoles can't show all 16 million colors.
     * This provides a visual test as to whether or not a terminal
     * can show all 16 million colors.
     *
     * If the small change in color scale is visible this means
     * the terminal is capable of outputting 16 million colors.
     * If the change is not visible IE most of it looks to be the
     * same shade then it is not capable.
     *
     * QUIRK: The console test fails on Clion's CMD Prompt (Windows). Windows CMD Prompt is fine
     */
    void printRGBConsoleTest()
    {
        auto redScale = RGBColorScale(Color4(), redScaler);
        auto greenScale = RGBColorScale(Color4(), greenScaler);
        auto blueScale = RGBColorScale(Color4(), blueScaler);

        printRGBScale(redScale);
        printRGBScale(greenScale);
        printRGBScale(blueScale);
    }
}